package com.friday.tooldriver.mapper.smartrpa;

import com.friday.tooldriver.dto.smartrpa.JobSelectFilter;
import com.friday.tooldriver.po.smartrpa.JobPo;
import org.springframework.stereotype.Repository;

import java.util.List;

@Repository
public interface JobExecMapper {
    void insertJob(JobPo job);
    void updateJobById(JobPo job);
    List<JobPo> selectJobs(JobSelectFilter filter);
    JobPo selectJobById(Integer jobId);
}
